@@ -4,17 +4,18 @@ static pe_ring IOWatch;
 static int IOWatchCount;
 static int IOWatch_OK;
 
-static pe_event *
-pe_io_allocate()
+static pe_event *pe_io_allocate()
 {
   pe_io *ev;
   New(PE_NEWID, ev, 1, pe_io);
   ev->base.vtbl = &pe_io_vtbl;
   pe_event_init((pe_event*) ev);
   PE_RING_INIT(&ev->tm.ring, ev);
+  PE_RING_INIT(&ev->ttm.ring, ev);
   PE_RING_INIT(&ev->ioring, ev);
   ev->fd = -1;
   ev->timeout = 0;
+  ev->tailpoll = 0;
   ev->handle = 0;
   ev->events = 0;
   ev->got = 0;
@@ -23,8 +24,7 @@ pe_io_allocate()
   return (pe_event*) ev;
 }
 
-static void
-pe_io_dtor(pe_event *_ev)
+static void pe_io_dtor(pe_event *_ev)
 {
   pe_io *ev = (pe_io*) _ev;
   PE_RING_DETACH(&ev->ioring);
@@ -33,63 +33,85 @@ pe_io_dtor(pe_event *_ev)
   (*_ev->vtbl->up->dtor)(_ev);
 }
 
-static SV*
-io_events_2sv(int mask)
-{
-  STRLEN len;
-  SV *ret = newSV(0);
-  SvUPGRADE(ret, SVt_PVIV);
-  sv_setpvn(ret, "", 0);
-  if (mask & PE_IO_R) sv_catpv(ret, "r");
-  if (mask & PE_IO_W) sv_catpv(ret, "w");
-  if (mask & PE_IO_E) sv_catpv(ret, "e");
-  if (mask & PE_IO_T) sv_catpv(ret, "t");
-  SvIVX(ret) = mask;
-  SvIOK_on(ret);
-  return ret;
-}
-
 static void pe_io_start(pe_event *_ev, int repeat)
 {
   pe_io *ev = (pe_io*) _ev;
+  ev->fd = pe_sys_fileno(ev);
   PE_RING_UNSHIFT(&ev->ioring, &IOWatch);
   ++IOWatchCount;
   IOWatch_OK = 0;
+  if (ev->tailpoll) {
+    struct stat sbuf;
+    int ret;
+    if (!(ev->events & PE_R))
+      croak("'tailpoll' causes read events but '%s' does not care about read",
+	    SvPV(_ev->desc,na));
+  retry:
+    ret = fstat(ev->fd, &sbuf);
+    if (ret) {
+      if (errno == EINTR) goto retry;
+      croak("Event '%s': fstat(%d) failed with %d", SvPV(_ev->desc,na),
+	    ev->fd, errno);
+    }
+    ev->size = sbuf.st_size;
+    ev->ttm.at = EvNOW(0) + ev->tailpoll;
+    pe_timeable_start(&ev->ttm);
+  }
   if (ev->timeout) {
     EvCBTIME_on(ev);
-    ev->events |= PE_IO_T;
+    ev->events |= PE_T;
     ev->tm.at = EvNOW(0) + ev->timeout;  /* too early okay */
-    pe_timeable_start(_ev);
+    pe_timeable_start(&ev->tm);
   } else {
     EvCBTIME_off(ev);
-    ev->events &= ~PE_IO_T;
+    ev->events &= ~PE_T;
   }
 }
 
 static void pe_io_stop(pe_event *_ev)
 {
   pe_io *ev = (pe_io*) _ev;
-  if (ev->timeout)
-    pe_timeable_stop(_ev);
+  pe_timeable_stop(&ev->tm);
+  pe_timeable_stop(&ev->ttm);
   PE_RING_DETACH(&ev->ioring);
   --IOWatchCount;
   IOWatch_OK = 0;
 }
 
-static void pe_io_alarm(pe_event *_ev)
+static void pe_io_alarm(pe_event *_ev, pe_timeable *hit)
 {
   pe_io *ev = (pe_io*) _ev;
   double now = EvNOW(1);
-  double left = (_ev->cbtime + ev->timeout) - now;
-  if (left < PE_INTERVAL_EPSILON) {
-    ev->tm.at = now + ev->timeout;
-    ev->got |= PE_IO_T;
-    queueEvent(_ev, 1);
+  if (hit == &ev->tm) {
+    double left = (_ev->cbtime + ev->timeout) - now;
+    if (left < PE_INTERVAL_EPSILON) {
+      ev->tm.at = now + ev->timeout;
+      ev->got |= PE_T;
+      queueEvent(_ev, 1);
+    }
+    else {
+      ev->tm.at = left;
+    }
+    pe_timeable_start(&ev->tm);
   }
   else {
-    ev->tm.at = left;
+    struct stat sbuf;
+    int ret;
+  retry:
+    ret = fstat(ev->fd, &sbuf);
+    if (ret) {
+      if (errno == EINTR) goto retry;
+      croak("Event '%s': fstat(%d) failed with %d", SvPV(_ev->desc,na),
+	    ev->fd, errno);
+    }
+    if (sbuf.st_size != ev->size) {
+      ev->size = sbuf.st_size;
+      ev->got |= PE_R;
+      queueEvent(_ev, 1);
+    }
+    ev->ttm.at = EvNOW(0) + ev->tailpoll;
+    pe_timeable_start(&ev->ttm);
   }
-  pe_timeable_start(_ev);
 }
 
 static void pe_io_FETCH(pe_event *_ev, SV *svkey)
@@ -124,6 +146,10 @@ static void pe_io_FETCH(pe_event *_ev, SV *svkey)
       ret = sv_2mortal(newSVnv(ev->timeout));
       break;
     }
+    if (len == 8 && memEQ(key, "tailpoll", 8)) {
+      ret = sv_2mortal(newSVnv(ev->tailpoll));
+      break;
+    }
     break;
   }
   if (ret) {
@@ -135,8 +161,7 @@ static void pe_io_FETCH(pe_event *_ev, SV *svkey)
   }
 }
 
-static void
-pe_io_STORE(pe_event *_ev, SV *svkey, SV *nval)
+static void pe_io_STORE(pe_event *_ev, SV *svkey, SV *nval)
 {
   pe_io *ev = (pe_io*) _ev;
   STRLEN len;
@@ -148,16 +173,16 @@ pe_io_STORE(pe_event *_ev, SV *svkey, SV *nval)
   case 'e':
     if (len == 6 && memEQ(key, "events", 6)) {
       ok=1;
-      ev->events = ev->timeout? PE_IO_T : 0;
+      ev->events = ev->timeout? PE_T : 0;
       if (SvPOK(nval)) {  /* can fall thru to SvIOK? XXX */
 	int xx;
 	STRLEN el;
 	char *ep = SvPV(nval,el);
 	for (xx=0; xx < el; xx++) {
 	  switch (ep[xx]) {
-	  case 'r': ev->events |= PE_IO_R; break;
-	  case 'w': ev->events |= PE_IO_W; break;
-	  case 'e': ev->events |= PE_IO_E; break;
+	  case 'r': ev->events |= PE_R; break;
+	  case 'w': ev->events |= PE_W; break;
+	  case 'e': ev->events |= PE_E; break;
 	  default: warn("ignored '%c'", ep[xx]);
 	  }
 	}
@@ -169,11 +194,11 @@ pe_io_STORE(pe_event *_ev, SV *svkey, SV *nval)
 	   want to switch to our own constants! */
 	mask = SvIV(nval);
 	if (mask & (POLLIN | POLLRDNORM))
-	  ev->events |= PE_IO_R;
+	  ev->events |= PE_R;
 	if (mask & (POLLOUT | POLLWRNORM | POLLWRBAND))
-	  ev->events |= PE_IO_W;
+	  ev->events |= PE_W;
 	if (mask & (POLLRDBAND | POLLPRI))
-	  ev->events |= PE_IO_E;
+	  ev->events |= PE_E;
 #endif
       } else
 	croak("Event::io::STORE('events'): expecting a string");
@@ -199,6 +224,11 @@ pe_io_STORE(pe_event *_ev, SV *svkey, SV *nval)
       ev->timeout = SvNV(nval);
       break;
     }
+    if (len == 8 && memEQ(key, "tailpoll", 8)) {
+      ok=1;
+      ev->tailpoll = SvNV(nval);
+      break;
+    }
     break;
   }
   if (ok) {
@@ -224,7 +254,9 @@ static void boot_io()
   static char *keylist[] = {
     "handle",
     "events",
-    "got"
+    "got",
+    "timeout",
+    "tailpoll"
   };
   HV *stash = (HV*) SvREFCNT_inc((SV*) gv_stashpv("Event::io",1));
   pe_event_vtbl *vt = &pe_io_vtbl;
@@ -240,10 +272,6 @@ static void boot_io()
   vt->stop = pe_io_stop;
   vt->postCB = pe_io_postCB;
   vt->alarm = pe_io_alarm;
-  newCONSTSUB(stash, "R", newSViv(PE_IO_R));
-  newCONSTSUB(stash, "W", newSViv(PE_IO_W));
-  newCONSTSUB(stash, "E", newSViv(PE_IO_E));
-  newCONSTSUB(stash, "T", newSViv(PE_IO_T));
   PE_RING_INIT(&IOWatch, 0);
   IOWatch_OK = 0;
   IOWatchCount = 0;
