@@ -2,12 +2,14 @@ static pe_ring NQueue;
 static pe_stat idleStats;
 static AV *Prepare, *Check, *AsyncCheck;
 static int StarvePrio = PE_QUEUES - 2;
+static double QueueTime[PE_QUEUES];
 
 static void boot_queue()
 {
   int xx;
   HV *stash = gv_stashpv("Event", 1);
   PE_RING_INIT(&NQueue, 0);
+  memset(QueueTime, 0, sizeof(QueueTime));
   newCONSTSUB(stash, "QUEUES", newSViv(PE_QUEUES));
   newCONSTSUB(stash, "PRIO_NORMAL", newSViv(PE_PRIO_NORMAL));
   newCONSTSUB(stash, "PRIO_HIGH", newSViv(PE_PRIO_HIGH));
@@ -42,11 +44,13 @@ static void db_show_queue()
   }
 }
 
-static void ready_event(pe_event *ev)
+static void prepare_event(pe_event *ev)
 {
   assert(!EvSUSPEND(ev));
-  assert(EvACTIVE(ev));    /* or how'd we get here?! */
   assert(EvREENTRANT(ev) || !ev->running);
+  if (!EvACTIVE(ev))
+    croak("Event: attempt to run callback for !ACTIVE watcher '%s'",
+	  SvPV(ev->desc,na));
 
   /* this cannot be done any later because the callback might want to
      again() or whatever */
@@ -56,7 +60,7 @@ static void ready_event(pe_event *ev)
 
 static void queueEvent(pe_event *ev, int count)
 {
-  ready_event(ev);
+  prepare_event(ev);
 
   assert(count >= 0);
   ev->count += count;
@@ -75,6 +79,7 @@ static void queueEvent(pe_event *ev, int count)
     ev->priority = PE_QUEUES-1;
   if (EvDEBUGx(ev) >= 3)
     warn("Event: queue '%s' prio=%d\n", SvPV(ev->desc,na), ev->priority);
+  QueueTime[ev->priority] = EvNOW(0);
   {
     /* queue in reverse direction? XXX */ 
     /*  warn("-- adding 0x%x/%d\n", ev, prio); db_show_queue();/**/
@@ -89,6 +94,33 @@ static void queueEvent(pe_event *ev, int count)
   }
 }
 
+static double pe_map_prepare(double tm)
+{
+  /* untested XXX */
+  int xx;
+  ENTER;
+  SAVETMPS;
+  for (xx=0; xx <= av_len(Prepare); xx++) {
+    SV *got;
+    SV **cv = av_fetch(Prepare, xx, 0);
+    dSP;
+    PUSHMARK(SP);
+    PUTBACK;
+    if (!cv) croak("$Prepare[xx] unset");
+    perl_call_sv(*cv, G_SCALAR);
+    SPAGAIN;
+    got = POPs;
+    PUTBACK;
+    if (SvOK(got) && SvNOK(got)) {
+      double when = SvNV(got);
+      if (when < tm) tm = when;
+    }
+  }
+  FREETMPS;
+  LEAVE;
+  return tm;
+}
+
 static void pe_map_check(AV *av)
 {
   int xx;
@@ -106,47 +138,60 @@ static void pe_map_check(AV *av)
   LEAVE;
 }
 
-/*
-  recover if exited via longjmp
+static int pe_empty_queue(maxprio)
+{
+  pe_event *ev = NQueue.next->self;
+  if (ev && ev->priority < maxprio) {
+    dequeEvent(ev);
+    if (EvDEBUGx(ev) >= 2)
+      warn("Event: invoking '%s' (prio %d)\n",
+	   SvPV(ev->desc, na), ev->priority);
+    pe_event_invoke(ev);
+    return 1;
+  }
+  return 0;
+}
 
-  @AsyncCheck
+/*inline*/ static void pe_multiplex(double tm)
+{
+  if (SvIVX(DebugLevel) >= 2) {
+    warn("Event: multiplex %.4fs %s%s\n", tm,
+	 PE_RING_EMPTY(&NQueue)?"":"QUEUE",
+	 PE_RING_EMPTY(&Idle)?"":"IDLE");
+  }
+  if (!Stats)
+    pe_sys_multiplex(tm);
+  else {
+    struct timeval start_tm, done_tm;
+    gettimeofday(&start_tm, 0);
+    pe_sys_multiplex(tm);
+    gettimeofday(&done_tm, 0);
+    pe_stat_record(&idleStats,
+		   (done_tm.tv_sec-start_tm.tv_sec +
+		    (done_tm.tv_usec-start_tm.tv_usec)/1000000.0));
+  }
+}
 
-  return 1 if emptyQueue(QUEUES-2)
+static void pe_queue_pending()
+{
+  double tm = 0;
+  if (av_len(Prepare) >= 0) tm = pe_map_prepare(tm);
 
-  tm = min @Prepare
+  pe_multiplex(0);
 
-  pe_sys_multiplex(tm)
+  pe_timeables_check();
+  if (av_len(Check) >= 0) pe_map_check(Check);
 
-  @Check
-
-  if (tm) @AsyncCheck
-
-  return 1 if emptyQueue(QUEUES)
-
-  return 1 runIdle
-
-  return 0
- */
-
-#define EMPTYQUEUE(MAX)				\
-STMT_START {					\
-  pe_event *ev = NQueue.next->self;		\
-  if (ev && ev->priority < MAX) {		\
-    dequeEvent(ev);				\
-    if (EvDEBUGx(ev) >= 2)			\
-      warn("Event: invoking '%s' (prio %d)\n",	\
-	   SvPV(ev->desc, na), ev->priority);	\
-    pe_event_invoke(ev);			\
-    return 1;					\
-  }						\
-} STMT_END
+  pe_signal_asynccheck();
+  if (av_len(AsyncCheck) >= 0) pe_map_check(AsyncCheck);
+}
 
 static int one_event(double tm)
 {
   pe_signal_asynccheck();
   if (av_len(AsyncCheck) >= 0) pe_map_check(AsyncCheck);
 
-  EMPTYQUEUE(StarvePrio);
+  if (pe_empty_queue(StarvePrio)) return 1;
 
   if (!PE_RING_EMPTY(&NQueue) || !PE_RING_EMPTY(&Idle)) {
     tm = 0;
@@ -155,46 +200,9 @@ static int one_event(double tm)
     double t1 = timeTillTimer();
     if (t1 < tm) tm = t1;
   }
-  if (av_len(Prepare) >= 0) {
-    /* untested XXX */
-    int xx;
-    ENTER;
-    SAVETMPS;
-    for (xx=0; xx <= av_len(Prepare); xx++) {
-      SV *got;
-      SV **cv = av_fetch(Prepare, xx, 0);
-      dSP;
-      PUSHMARK(SP);
-      PUTBACK;
-      if (!cv) croak("$Prepare[xx] unset");
-      perl_call_sv(*cv, G_SCALAR);
-      SPAGAIN;
-      got = POPs;
-      PUTBACK;
-      if (SvOK(got) && SvNOK(got)) {
-	double when = SvNV(got);
-	if (when < tm) tm = when;
-      }
-    }
-    FREETMPS;
-    LEAVE;
-  }
+  if (av_len(Prepare) >= 0) tm = pe_map_prepare(tm);
 
-  if (SvIVX(DebugLevel) >= 2) {
-    warn("Event: multiplex %.4fs %s%s\n", tm,
-	 PE_RING_EMPTY(&NQueue)?"":"QUEUE",
-	 PE_RING_EMPTY(&Idle)?"":"IDLE");
-  }
-  if (!Stats)
-    pe_sys_multiplex(tm);
-  else {  
-    struct timeval start_tm, done_tm;
-    gettimeofday(&start_tm, 0);
-    pe_sys_multiplex(tm);
-    gettimeofday(&done_tm, 0);
-    pe_stat_record(&idleStats, (done_tm.tv_sec-start_tm.tv_sec +
-				(done_tm.tv_usec-start_tm.tv_usec)/1000000.0));
-  }
+  pe_multiplex(tm);
 
   pe_timeables_check();
   if (av_len(Check) >= 0) pe_map_check(Check);
@@ -204,13 +212,13 @@ static int one_event(double tm)
     if (av_len(AsyncCheck) >= 0) pe_map_check(AsyncCheck);
   }
 
-  EMPTYQUEUE(PE_QUEUES);
+  if (pe_empty_queue(PE_QUEUES)) return 1;
 
   {
     pe_event *ev;
     if (PE_RING_EMPTY(&Idle)) return 0;
     PE_RING_POP(&Idle, ev);
-    ready_event(ev);
+    prepare_event(ev);
     /* can't queueEvent because we are already beyond that */
     ++ev->count;
     if (EvDEBUGx(ev) >= 2)
@@ -234,82 +242,3 @@ static void pe_unloop(SV *why)
   sv_setsv(result, why);
   sv_dec(exitL);
 }
-
-typedef struct pe_sleep_frame pe_sleep_frame;
-struct pe_sleep_frame {
-  int expired;
-  int was_reentrant;
-  pe_event *cb;
-  pe_event *tmr;
-  SV *ret;
-};
-
-static void pe_sleep_expire(void *vptr)
-{
-  pe_sleep_frame *fr = (pe_sleep_frame *) vptr;
-  fr->expired = 1;
-  pe_unloop(fr->ret);
-}
-
-static SV *pe_sleep(SV *howlong)
-{
-  double duration;
-  if (!sv_2interval(howlong, &duration))
-    croak("Event::sleep expecting a duration");
-  if (CurCBFrame < 0) {
-    if (SvIVX(DebugLevel) >= 3)
-      warn("Event: sleep(%.2f)\n", duration);
-    pe_sys_sleep(duration);
-    return howlong;
-  }
-  else {
-    pe_sleep_frame fr;
-    pe_timer *tm;
-    SV *ret;
-
-    fr.expired = 0;
-    fr.tmr = pe_timer_allocate();
-    fr.cb = (CBFrame + CurCBFrame)->ev;
-    fr.ret = SvREFCNT_inc(howlong);
-
-    /* protect callback event */
-    ++fr.cb->refcnt;
-    fr.was_reentrant = EvREENTRANT(fr.cb);
-    EvREENTRANT_off(fr.cb);
-
-    /* set up timer */
-    tm = (pe_timer*) fr.tmr;
-    ++tm->base.refcnt;
-    sv_setpvf(tm->base.desc, "sleep(%.2f id=%d) timer", duration, fr.cb->id);
-    tm->tm.at = EvNOW(1) + duration;
-    tm->base.c_callback = pe_sleep_expire;
-    tm->base.ext_data = &fr;
-    tm->base.priority = PE_PRIO_HIGH;
-    pe_event_start((pe_event*) tm, 0);
-
-    if (SvIVX(DebugLevel) >= 2)
-      warn("Event: sleep(%.2f) id=%d\n", duration, fr.cb->id);
-
-    {
-      dSP;
-      PUSHMARK(SP);
-      perl_call_pv("Event::loop", G_SCALAR | G_NOARGS);  /* cannot die */
-      SPAGAIN;
-      ret = POPs;
-      PUTBACK;
-    }
-
-    if (SvIVX(DebugLevel) >= 2)
-      warn("Event: sleep id=%d %s\n", fr.cb->id,
-	   fr.expired? "finished":"unlooped early");
-    if (fr.was_reentrant)
-      EvREENTRANT_on(fr.cb);
-    --fr.cb->refcnt;
-    --fr.tmr->refcnt;
-    sv_2mortal(howlong);
-    pe_event_cancel(fr.tmr);
-
-    return ret;
-  }
-}
-
